@@ -160,11 +160,16 @@ void     _gdk_moveresize_handle_event      (XEvent *event);
 void     _gdk_moveresize_configure_done    (void);
 
 void     _gdk_keymap_state_changed         (void);
+gint     _gdk_x11_get_group_for_state      (GdkModifierType state);
 
 GC _gdk_x11_gc_flush (GdkGC *gc);
 
 void _gdk_x11_initialize_locale (void);
 
+void _gdk_xgrab_check_unmap   (GdkWindow *window,
+			       gulong     serial);
+void _gdk_xgrab_check_destroy (GdkWindow *window);
+
 extern GdkDrawableClass  _gdk_x11_drawable_class;
 extern Window		 _gdk_root_window;
 extern gboolean	         _gdk_use_xshm;
@@ -177,10 +182,6 @@ extern gchar		*_gdk_display_name;
 
 extern Window		 _gdk_leader_window;
 
-extern GdkWindowObject *_gdk_xgrab_window;  /* Window that currently holds the
-					    * x pointer grab
-					    */
-
 /* Used to detect not-up-to-date keymap */
 extern guint _gdk_keymap_serial;
 
