@@ -748,8 +748,12 @@ gtk_menu_popup (GtkMenu		    *menu,
    */
   gtk_widget_show (GTK_WIDGET (menu));
 
+  /* Position the menu, possibly changing the size request
+   */
+  gtk_menu_position (menu);
+
   /* Compute the size of the toplevel and realize it so we
-   * can position and scroll correctly.
+   * can scroll correctly.
    */
   {
     GtkRequisition tmp_request;
@@ -765,8 +769,6 @@ gtk_menu_popup (GtkMenu		    *menu,
     gtk_widget_realize (GTK_WIDGET (menu));
   }
 
-  gtk_menu_position (menu);
-
   gtk_menu_scroll_to (menu, menu->scroll_offset);
 
   /* Once everything is set up correctly, map the toplevel window on
@@ -1110,13 +1112,6 @@ gtk_menu_set_tearoff_state (GtkMenu  *menu,
 
 	      gtk_widget_realize (menu->tearoff_window);
 	      
-	      gdk_window_set_decorations (menu->tearoff_window->window, 
-					  GDK_DECOR_ALL |
-					  GDK_DECOR_RESIZEH |
-					  GDK_DECOR_MINIMIZE |
-					  GDK_DECOR_MAXIMIZE);
-	      gtk_window_set_resizable (GTK_WINDOW (menu->tearoff_window), FALSE);
-
 	      menu->tearoff_hbox = gtk_hbox_new (FALSE, FALSE);
 	      gtk_container_add (GTK_CONTAINER (menu->tearoff_window), menu->tearoff_hbox);
 
@@ -1665,35 +1660,6 @@ gtk_menu_paint (GtkWidget      *widget,
 			   MENU_SCROLL_ARROW_HEIGHT - 2 * border_y - 2);
 	}
     }
-  else if (event->window == menu->view_window)
-    {
-      gint menu_height;
-      gint top_pos;
-      
-      if (menu->scroll_offset < 0)
-	gtk_paint_box (widget->style,
-		       menu->view_window,
-		       GTK_STATE_ACTIVE,
-		       GTK_SHADOW_IN,
-		       NULL, widget, "menu",
-		       0, 0,
-		       -1,
-		       -menu->scroll_offset);
-
-      menu_height = widget->requisition.height - 2*border_y;
-      top_pos = height - 2*border_y - (menu->upper_arrow_visible ? MENU_SCROLL_ARROW_HEIGHT : 0);
-
-      if (menu_height - menu->scroll_offset < top_pos)
-	gtk_paint_box (widget->style,
-		       menu->view_window,
-		       GTK_STATE_ACTIVE,
-		       GTK_SHADOW_IN,
-		       NULL, widget, "menu",
-		       0,
-		       menu_height - menu->scroll_offset,
-		       -1,
-		       top_pos - (menu_height - menu->scroll_offset));
-    }
 }
 
 static gboolean
@@ -1733,6 +1699,7 @@ gtk_menu_key_press (GtkWidget	*widget,
   gboolean can_change_accels;
   gchar *accel = NULL;
   guint accel_key, accel_mods;
+  GdkModifierType consumed_modifiers;
   
   g_return_val_if_fail (GTK_IS_MENU (widget), FALSE);
   g_return_val_if_fail (event != NULL, FALSE);
@@ -1791,16 +1758,27 @@ gtk_menu_key_press (GtkWidget	*widget,
   g_object_get (G_OBJECT (gtk_settings_get_default ()),
 		"gtk-can-change-accels", &can_change_accels,
 		NULL);
-  
-  accel_key = event->keyval;
-  accel_mods = event->state & gtk_accelerator_get_default_mod_mask ();
 
+  /* Figure out what modifiers went into determining the key symbol */
+  gdk_keymap_translate_keyboard_state (gdk_keymap_get_default (),
+				       event->hardware_keycode, event->state, event->group,
+				       NULL, NULL, NULL, &consumed_modifiers);
+
+  accel_key = gdk_keyval_to_lower (event->keyval);
+  accel_mods = event->state & gtk_accelerator_get_default_mod_mask () & ~consumed_modifiers;
+
+  /* If lowercasing affects the keysym, then we need to include SHIFT in the modifiers,
+   * We re-upper case when we match against the keyval, but display and save in caseless form.
+   */
+  if (accel_key != event->keyval)
+    accel_mods |= GDK_SHIFT_MASK;
+  
   /* Modify the accelerators */
   if (can_change_accels &&
       menu_shell->active_menu_item &&
       GTK_BIN (menu_shell->active_menu_item)->child &&			/* no seperators */
       GTK_MENU_ITEM (menu_shell->active_menu_item)->submenu == NULL &&	/* no submenus */
-      (delete || gtk_accelerator_valid (event->keyval, event->state)))
+      (delete || gtk_accelerator_valid (accel_key, accel_mods)))
     {
       GtkWidget *menu_item = menu_shell->active_menu_item;
       gboolean replace_accels = TRUE;
@@ -1860,7 +1838,6 @@ gtk_menu_motion_notify  (GtkWidget	   *widget,
 
   gboolean need_enter;
 
-
   if (GTK_IS_MENU (widget))
     gtk_menu_handle_scrolling (GTK_MENU (widget), TRUE);
   
@@ -2122,7 +2099,7 @@ gtk_menu_leave_notify (GtkWidget        *widget,
       && menu_item->submenu != NULL
       && menu_item->submenu_placement == GTK_LEFT_RIGHT)
     {
-      if (menu_item->submenu->window != NULL) 
+      if (GTK_MENU_SHELL (menu_item->submenu)->active)
 	{
 	  gtk_menu_set_submenu_navigation_region (menu, menu_item, event);
 	  return TRUE;
@@ -2242,7 +2219,7 @@ gtk_menu_set_submenu_navigation_region (GtkMenu          *menu,
       /* Exiting the top or bottom? */ 
       if (event->y < 0)
         { /* top */
-	  point[0].y = event->y_root + SUBMENU_NAV_REGION_PADDING;
+	  point[0].y = event->y_root + 1;
 	  point[1].y = submenu_top;
 
 	  if (point[0].y <= point[1].y)
@@ -2250,7 +2227,7 @@ gtk_menu_set_submenu_navigation_region (GtkMenu          *menu,
 	}
       else
         { /* bottom */
-	  point[0].y = event->y_root - SUBMENU_NAV_REGION_PADDING; 
+	  point[0].y = event->y_root;
 	  point[1].y = submenu_bottom;
 
 	  if (point[0].y >= point[1].y)
