@@ -30,7 +30,8 @@ enum {
   PROP_THEME_NAME,
   PROP_KEY_THEME_NAME,
   PROP_MENU_BAR_ACCEL,
-  PROP_DND_DRAG_THRESHOLD
+  PROP_DND_DRAG_THRESHOLD,
+  PROP_FONT_NAME
 };
 
 
@@ -178,7 +179,7 @@ gtk_settings_class_init (GtkSettingsClass *class)
                                              g_param_spec_string ("gtk-theme-name",
 								   _("Theme Name"),
 								   _("Name of theme RC file to load"),
-								  NULL,
+								  "Default",
 								  G_PARAM_READWRITE),
                                              NULL);
   g_assert (result == PROP_THEME_NAME);
@@ -208,7 +209,16 @@ gtk_settings_class_init (GtkSettingsClass *class)
                                                                G_PARAM_READWRITE),
 					     NULL);
   g_assert (result == PROP_DND_DRAG_THRESHOLD);
-  
+
+  result = settings_install_property_parser (class,
+                                             g_param_spec_string ("gtk-font-name",
+								   _("Font Name"),
+								   _("Name of default font to use"),
+								  "Sans 10",
+								  G_PARAM_READWRITE),
+                                             NULL);
+  g_assert (result == PROP_FONT_NAME);
+ 
 }
 
 static void
@@ -309,8 +319,7 @@ gtk_settings_get_property (GObject     *object,
         {
           GValue tmp_value = { 0, };
           GValue gstring_value = { 0, };
-          GtkRcPropertyParser parser = g_param_spec_get_qdata (pspec,
-                                                               quark_property_parser);
+          GtkRcPropertyParser parser = (GtkRcPropertyParser) g_param_spec_get_qdata (pspec, quark_property_parser);
           
           g_value_init (&gstring_value, G_TYPE_GSTRING);
 
@@ -438,7 +447,7 @@ apply_queued_setting (GtkSettings      *data,
 		      GtkSettingsValue *qvalue)
 {
   GValue tmp_value = { 0, };
-  GtkRcPropertyParser parser = g_param_spec_get_qdata (pspec, quark_property_parser);
+  GtkRcPropertyParser parser = (GtkRcPropertyParser) g_param_spec_get_qdata (pspec, quark_property_parser);
 
   g_value_init (&tmp_value, G_PARAM_SPEC_VALUE_TYPE (pspec));
   if (_gtk_settings_parse_convert (parser, &qvalue->value,
@@ -498,7 +507,7 @@ settings_install_property_parser (GtkSettingsClass   *class,
     g_object_freeze_notify (node->data);
 
   g_object_class_install_property (G_OBJECT_CLASS (class), ++class_n_properties, pspec);
-  g_param_spec_set_qdata (pspec, quark_property_parser, parser);
+  g_param_spec_set_qdata (pspec, quark_property_parser, (gpointer) parser);
 
   for (node = object_list; node; node = node->next)
     {
