@@ -64,6 +64,7 @@ struct _GtkTreeViewColumnCellInfo
   gpointer func_data;
   GtkDestroyNotify destroy;
   gint requested_width;
+  gint real_width;
   guint expand : 1;
   guint pack : 1;
   guint has_focus : 1;
@@ -329,8 +330,14 @@ gtk_tree_view_column_finalize (GObject *object)
   for (list = tree_column->cell_list; list; list = list->next)
     {
       GtkTreeViewColumnCellInfo *info = (GtkTreeViewColumnCellInfo *) list->data;
-      if (info->func_data && info->destroy)
-	(info->destroy) (info->func_data);
+
+      if (info->destroy)
+	{
+	  GtkDestroyNotify d = info->destroy;
+
+	  info->destroy = NULL;
+	  d (info->func_data);
+	}
       gtk_tree_view_column_clear_attributes (tree_column, info->cell);
       g_object_unref (G_OBJECT (info->cell));
       g_free (info);
@@ -542,7 +549,6 @@ gtk_tree_view_column_create_button (GtkTreeViewColumn *tree_column)
   g_signal_connect (G_OBJECT (tree_column->button), "event",
 		    G_CALLBACK (gtk_tree_view_column_button_event),
 		    (gpointer) tree_column);
-
   g_signal_connect (G_OBJECT (tree_column->button), "clicked",
 		    (GtkSignalFunc) gtk_tree_view_column_button_clicked,
 		    (gpointer) tree_column);
@@ -655,10 +661,10 @@ gtk_tree_view_column_update_button (GtkTreeViewColumn *tree_column)
       break;
     }
 
-  /* Put arrow on the right if the text is left-or-center justified,
-       * and on the left otherwise; do this by packing boxes, so flipping
-       * text direction will reverse things
-       */
+  /* Put arrow on the right if the text is left-or-center justified, and on the
+   * left otherwise; do this by packing boxes, so flipping text direction will
+   * reverse things
+   */
   gtk_widget_ref (arrow);
   gtk_container_remove (GTK_CONTAINER (hbox), arrow);
 
@@ -679,8 +685,8 @@ gtk_tree_view_column_update_button (GtkTreeViewColumn *tree_column)
   else
     gtk_widget_hide (arrow);
 
-  /* It's always safe to hide the button.  It isn't always safe to show it, as if you show it
-   * before it's realized, it'll get the wrong window. */
+  /* It's always safe to hide the button.  It isn't always safe to show it, as
+   * if you show it before it's realized, it'll get the wrong window. */
   if (tree_column->button &&
       tree_column->tree_view != NULL &&
       GTK_WIDGET_REALIZED (tree_column->tree_view))
@@ -811,17 +817,13 @@ gtk_tree_view_column_mnemonic_activate (GtkWidget *widget,
 
   g_return_val_if_fail (GTK_IS_TREE_VIEW_COLUMN (column), FALSE);
 
+  GTK_TREE_VIEW (column->tree_view)->priv->focus_column = column;
   if (column->clickable)
     gtk_button_clicked (GTK_BUTTON (column->button));
   else if (GTK_WIDGET_CAN_FOCUS (column->button))
     gtk_widget_grab_focus (column->button);
   else
-    {
-      GTK_TREE_VIEW (column->tree_view)->priv->focus_column = column;
-      GTK_TREE_VIEW_SET_FLAG (GTK_TREE_VIEW (column->tree_view),
-			      GTK_TREE_VIEW_DRAW_KEYFOCUS);
-      gtk_widget_grab_focus (column->tree_view);
-    }
+    gtk_widget_grab_focus (column->tree_view);
 
   return TRUE;
 }
@@ -1039,6 +1041,50 @@ _gtk_tree_view_column_unset_tree_view (GtkTreeViewColumn *column)
   column->button = NULL;
 }
 
+gboolean
+_gtk_tree_view_column_has_editable_cell (GtkTreeViewColumn *column)
+{
+  GList *list;
+
+  for (list = column->cell_list; list; list = list->next)
+    if (((GtkTreeViewColumnCellInfo *)list->data)->cell->mode ==
+	GTK_CELL_RENDERER_MODE_EDITABLE)
+      return TRUE;
+
+  return FALSE;
+}
+
+GtkCellRenderer *
+_gtk_tree_view_column_get_editable_cell (GtkTreeViewColumn *column)
+{
+  GList *list;
+
+  for (list = column->cell_list; list; list = list->next)
+    if (((GtkTreeViewColumnCellInfo *)list->data)->cell->mode ==
+	GTK_CELL_RENDERER_MODE_EDITABLE)
+      return ((GtkTreeViewColumnCellInfo *)list->data)->cell;
+
+  return NULL;
+}
+
+gint
+_gtk_tree_view_column_count_special_cells (GtkTreeViewColumn *column)
+{
+  gint i = 0;
+  GList *list;
+
+  for (list = column->cell_list; list; list = list->next)
+    {
+      GtkTreeViewColumnCellInfo *cellinfo = list->data;
+
+      if (cellinfo->cell->mode == GTK_CELL_RENDERER_MODE_EDITABLE ||
+	  cellinfo->cell->mode == GTK_CELL_RENDERER_MODE_ACTIVATABLE)
+	i++;
+    }
+
+  return i;
+}
+
 /* Public Functions */
 
 
@@ -1069,6 +1115,22 @@ gtk_tree_view_column_new (void)
  * equivalent to calling gtk_tree_view_column_set_title(),
  * gtk_tree_view_column_pack_start(), and
  * gtk_tree_view_column_set_attributes() on the newly created #GtkTreeViewColumn.
+ *
+ * Here's a simple example:
+ * <informalexample><programlisting>
+ *  <!>enum { TEXT_COLUMN, COLOR_COLUMN, N_COLUMNS };
+ *  <!>...
+ *  <!>{
+ *  <!>  GtkTreeViewColumn *column;
+ *  <!>  GtkCellRenderer   *renderer = gtk_cell_renderer_text_new ();
+ *  <!>
+ *  <!>  column = gtk_tree_view_column_new_with_attributes ("Title",
+ *  <!>                                                     renderer,
+ *  <!>                                                     "text", TEXT_COLUMN,
+ *  <!>                                                     "foreground", COLOR_COLUMN,
+ *  <!>                                                     NULL);
+ *  <!>}
+ * </programlisting></informalexample>
  * 
  * Return value: A newly created #GtkTreeViewColumn.
  **/
@@ -1160,7 +1222,7 @@ gtk_tree_view_column_pack_end (GtkTreeViewColumn  *tree_column,
   g_object_ref (G_OBJECT (cell));
   gtk_object_sink (GTK_OBJECT (cell));
 
-  cell_info = g_new (GtkTreeViewColumnCellInfo, 1);
+  cell_info = g_new0 (GtkTreeViewColumnCellInfo, 1);
   cell_info->cell = cell;
   cell_info->expand = expand ? TRUE : FALSE;
   cell_info->pack = GTK_PACK_END;
@@ -1187,8 +1249,8 @@ gtk_tree_view_column_clear (GtkTreeViewColumn *tree_column)
     {
       GtkTreeViewColumnCellInfo *info = (GtkTreeViewColumnCellInfo *)list->data;
 
-      g_object_unref (G_OBJECT (info->cell));
       gtk_tree_view_column_clear_attributes (tree_column, info->cell);
+      g_object_unref (G_OBJECT (info->cell));
       g_free (info);
     }
 
@@ -1333,13 +1395,13 @@ gtk_tree_view_column_set_cell_data_func (GtkTreeViewColumn   *tree_column,
 
   g_return_if_fail (info != NULL);
 
-  if (func == info->func &&
-      func_data == info->func_data &&
-      destroy == info->destroy)
-    return;
+  if (info->destroy)
+    {
+      GtkDestroyNotify d = info->destroy;
 
-  if (info->func_data && info->destroy)
-    (info->destroy) (info->func_data);
+      info->destroy = NULL;
+      d (info->func_data);
+    }
 
   info->func = func;
   info->func_data = func_data;
@@ -1445,6 +1507,9 @@ gtk_tree_view_column_set_visible (GtkTreeViewColumn *tree_column,
 
   tree_column->visible = visible;
 
+  if (tree_column->visible)
+    _gtk_tree_view_column_cell_set_dirty (tree_column, TRUE);
+
   gtk_tree_view_column_update_button (tree_column);
   g_object_notify (G_OBJECT (tree_column), "visible");
 }
@@ -1472,9 +1537,9 @@ gtk_tree_view_column_get_visible (GtkTreeViewColumn *tree_column)
  * @resizable: %TRUE, if the column can be resized
  * 
  * If @resizable is %TRUE, then the user can explicitly resize the column by
- * grabbing the outer edge of the column button.  If the sizing mode of the
- * column is #GTK_TREE_VIEW_COLUMN_AUTOSIZE, then it is changed to
- * #GTK_TREE_VIEW_COLUMN_GROW_ONLY.
+ * grabbing the outer edge of the column button.  If resizable is TRUE and
+ * sizing mode of the column is #GTK_TREE_VIEW_COLUMN_AUTOSIZE, then the sizing
+ * mode is changed to #GTK_TREE_VIEW_COLUMN_GROW_ONLY.
  **/
 void
 gtk_tree_view_column_set_resizable (GtkTreeViewColumn *tree_column,
@@ -1588,9 +1653,10 @@ gtk_tree_view_column_get_width (GtkTreeViewColumn *tree_column)
  * @fixed_width: The size to set @tree_column to. Must be greater than 0.
  * 
  * Sets the size of the column in pixels.  This is meaningful only if the sizing
- * type is #GTK_TREE_VIEW_COLUMN_FIXED.  In this case, the value is discarded
- * as the size of the column is based on the calculated width of the column. The
- * width is clamped to the min/max width for the column.
+ * type is #GTK_TREE_VIEW_COLUMN_FIXED.  The size of the column is clamped to
+ * the min/max width for the column.  Please note that the min/max width of the
+ * column doesn't actually affect the "fixed_width" property of the widget, just
+ * the actual size when displayed.
  **/
 void
 gtk_tree_view_column_set_fixed_width (GtkTreeViewColumn *tree_column,
@@ -1599,9 +1665,6 @@ gtk_tree_view_column_set_fixed_width (GtkTreeViewColumn *tree_column,
   g_return_if_fail (GTK_IS_TREE_VIEW_COLUMN (tree_column));
   g_return_if_fail (fixed_width > 0);
 
-  if (tree_column->column_type == GTK_TREE_VIEW_COLUMN_AUTOSIZE)
-    return;
-
   tree_column->fixed_width = fixed_width;
 
   if (tree_column->tree_view &&
@@ -1666,6 +1729,10 @@ gtk_tree_view_column_set_min_width (GtkTreeViewColumn *tree_column,
     }
   g_object_notify (G_OBJECT (tree_column), "min_width");
   g_object_thaw_notify (G_OBJECT (tree_column));
+
+  if (tree_column->column_type == GTK_TREE_VIEW_COLUMN_AUTOSIZE)
+    _gtk_tree_view_column_autosize (GTK_TREE_VIEW (tree_column->tree_view),
+				    tree_column);
 }
 
 /**
@@ -1722,6 +1789,10 @@ gtk_tree_view_column_set_max_width (GtkTreeViewColumn *tree_column,
     }
   g_object_notify (G_OBJECT (tree_column), "max_width");
   g_object_thaw_notify (G_OBJECT (tree_column));
+
+  if (tree_column->column_type == GTK_TREE_VIEW_COLUMN_AUTOSIZE)
+    _gtk_tree_view_column_autosize (GTK_TREE_VIEW (tree_column->tree_view),
+				    tree_column);
 }
 
 /**
@@ -1871,11 +1942,10 @@ gtk_tree_view_column_set_widget (GtkTreeViewColumn *tree_column,
  * gtk_tree_view_column_get_widget:
  * @tree_column: A #GtkTreeViewColumn.
  * 
- * Returns the #GtkWidget in the button in the column header.  If a custom
- * widget has not been set, then this will be a #GtkAlignment with a #GtkLabel
- * in it.
+ * Returns the #GtkWidget in the button on the column header.  If a custom
+ * widget has not been set then %NULL is returned.
  * 
- * Return value: The #GtkWidget in the column header.
+ * Return value: The #GtkWidget in the column header, or %NULL
  **/
 GtkWidget *
 gtk_tree_view_column_get_widget (GtkTreeViewColumn *tree_column)
@@ -2245,7 +2315,7 @@ gtk_tree_view_column_cell_get_size (GtkTreeViewColumn *tree_column,
       info->requested_width = MAX (info->requested_width, new_width + focus_line_width * 2);
       if (width)
 	* width += info->requested_width;
-      first_cell = TRUE;
+      first_cell = FALSE;
     }
 }
 
@@ -2253,15 +2323,24 @@ gtk_tree_view_column_cell_get_size (GtkTreeViewColumn *tree_column,
  * code.  Rather than duplicate them, we put them together to keep the code in
  * one place
  */
-static void
-gtk_tree_view_column_cell_render_or_focus (GtkTreeViewColumn *tree_column,
-					   GdkWindow         *window,
-					   GdkRectangle      *background_area,
-					   GdkRectangle      *cell_area,
-					   GdkRectangle      *expose_area,
-					   guint              flags,
-					   gboolean           render,
-					   GdkRectangle      *focus_rectangle)
+enum {
+  CELL_ACTION_RENDER,
+  CELL_ACTION_FOCUS,
+  CELL_ACTION_EVENT
+};
+
+static gboolean
+gtk_tree_view_column_cell_process_action (GtkTreeViewColumn  *tree_column,
+					  GdkWindow          *window,
+					  GdkRectangle       *background_area,
+					  GdkRectangle       *cell_area,
+					  guint               flags,
+					  gint                action,
+					  GdkRectangle       *expose_area,     /* RENDER */
+					  GdkRectangle       *focus_rectangle, /* FOCUS  */
+					  GtkCellEditable   **editable_widget, /* EVENT  */
+					  GdkEvent           *event,           /* EVENT  */
+					  gchar              *path_string)     /* EVENT  */
 {
   GList *list;
   GdkRectangle real_cell_area;
@@ -2279,7 +2358,8 @@ gtk_tree_view_column_cell_render_or_focus (GtkTreeViewColumn *tree_column,
   real_cell_area = *cell_area;
 
   gtk_widget_style_get (GTK_WIDGET (tree_column->tree_view),
-			"focus-line-width", &focus_line_width, NULL);
+			"focus-line-width", &focus_line_width,
+			NULL);
   /* Find out how my extra space we have to allocate */
   for (list = tree_column->cell_list; list; list = list->next)
     {
@@ -2309,8 +2389,9 @@ gtk_tree_view_column_cell_render_or_focus (GtkTreeViewColumn *tree_column,
 
       real_cell_area.width = info->requested_width +
 	(info->expand?extra_space:0);
+      info->real_width = real_cell_area.width;
       real_cell_area.x += focus_line_width;
-      if (render)
+      if (action == CELL_ACTION_RENDER)
 	{
 	  gtk_cell_renderer_render (info->cell,
 				    window,
@@ -2320,7 +2401,7 @@ gtk_tree_view_column_cell_render_or_focus (GtkTreeViewColumn *tree_column,
 				    expose_area,
 				    flags);
 	}
-      else
+      else if (action == CELL_ACTION_FOCUS)
 	{
 	  gint x_offset, y_offset, width, height;
 
@@ -2339,6 +2420,70 @@ gtk_tree_view_column_cell_render_or_focus (GtkTreeViewColumn *tree_column,
 	  if (max_y < real_cell_area.y + y_offset + height)
 	    max_y = real_cell_area.y + y_offset + height;
 	}
+      else if (action == CELL_ACTION_EVENT)
+	{
+	  gboolean try_event = FALSE;
+
+	  if (event)
+	    {
+	      if (_gtk_tree_view_column_count_special_cells (tree_column) == 1)
+	        {
+		  /* only 1 activatable cell -> whole column can activate */
+		  if (cell_area->x <= ((GdkEventButton *)event)->x &&
+		      cell_area->x + cell_area->width > ((GdkEventButton *)event)->x)
+		    try_event = TRUE;
+		}
+	      else if (real_cell_area.x <= ((GdkEventButton *)event)->x &&
+		  real_cell_area.x + real_cell_area.width > ((GdkEventButton *)event)->x)
+		  /* only activate cell if the user clicked on an individual
+		   * cell
+		   */
+		try_event = TRUE;
+	    }
+	  else /* if (info->has_focus)*/
+	    /* FIXME 73676: allow focusing individual cells (fixed on HEAD) */
+	    {
+	      try_event = TRUE;
+	    }
+	  if (try_event)
+	    {
+	      gboolean visible, mode;
+
+	      g_object_get (G_OBJECT (info->cell),
+			    "visible", &visible,
+			    "mode", &mode,
+			    NULL);
+	      if (visible && mode == GTK_CELL_RENDERER_MODE_ACTIVATABLE)
+		{
+		  if (gtk_cell_renderer_activate (info->cell,
+						  event,
+						  tree_column->tree_view,
+						  path_string,
+						  background_area,
+						  cell_area,
+						  flags))
+		    return TRUE;
+		}
+	      else if (visible && mode == GTK_CELL_RENDERER_MODE_EDITABLE)
+		{
+		  *editable_widget =
+		    gtk_cell_renderer_start_editing (info->cell,
+						     event,
+						     tree_column->tree_view,
+						     path_string,
+						     background_area,
+						     cell_area,
+						     flags);
+
+		  if (*editable_widget != NULL)
+		    {
+		      g_return_val_if_fail (GTK_IS_CELL_EDITABLE (*editable_widget), FALSE);
+		      
+		      return TRUE;
+		    }
+		}
+	    }
+	}
       real_cell_area.x += (info->requested_width + tree_column->spacing);
     }
   for (list = g_list_last (tree_column->cell_list); list; list = list->prev)
@@ -2353,24 +2498,110 @@ gtk_tree_view_column_cell_render_or_focus (GtkTreeViewColumn *tree_column,
 
       real_cell_area.width = info->requested_width +
 	(info->expand?extra_space:0);
-      gtk_cell_renderer_render (info->cell,
-				window,
-				tree_column->tree_view,
-				background_area,
-				&real_cell_area,
-				expose_area,
-				flags);
+      info->real_width = real_cell_area.width;
+      if (action == CELL_ACTION_RENDER)
+	{
+	  gtk_cell_renderer_render (info->cell,
+				    window,
+				    tree_column->tree_view,
+				    background_area,
+				    &real_cell_area,
+				    expose_area,
+				    flags);
+	}
+      else if (action == CELL_ACTION_FOCUS)
+	{
+	  gint x_offset, y_offset, width, height;
+
+	  gtk_cell_renderer_get_size (info->cell,
+				      tree_column->tree_view,
+				      &real_cell_area,
+				      &x_offset, &y_offset,
+				      &width, &height);
+
+	  if (min_x > (real_cell_area.x + x_offset))
+	    min_x = real_cell_area.x + x_offset;
+	  if (max_x < real_cell_area.x + x_offset + width)
+	    max_x = real_cell_area.x + x_offset + width;
+	  if (min_y > (real_cell_area.y + y_offset))
+	    min_y = real_cell_area.y + y_offset;
+	  if (max_y < real_cell_area.y + y_offset + height)
+	    max_y = real_cell_area.y + y_offset + height;
+	}
+      else if (action == CELL_ACTION_EVENT)
+        {
+	  gboolean try_event = FALSE;
+
+	  if (event)
+	    {
+	      if (_gtk_tree_view_column_count_special_cells (tree_column) == 1)
+	        {
+		  /* only 1 activatable cell -> whole column can activate */
+		  if (cell_area->x <= ((GdkEventButton *)event)->x &&
+		      cell_area->x + cell_area->width > ((GdkEventButton *)event)->x)
+		    try_event = TRUE;
+		}
+	      else if (real_cell_area.x <= ((GdkEventButton *)event)->x &&
+		  real_cell_area.x + real_cell_area.width > ((GdkEventButton *)event)->x)
+		/* only activate cell if the user clicked on an individual
+		 * cell
+		 */
+		try_event = TRUE;
+	    }
+	  else /* if (info->has_focus) */
+	    /* FIXME 73676: allow focusing individual cells (fixed on HEAD) */
+	    {
+	      try_event = TRUE;
+	    }
+	  if (try_event)
+	    {
+	      gboolean visible, mode;
+
+	      g_object_get (G_OBJECT (info->cell),
+			    "visible", &visible,
+			    "mode", &mode,
+			    NULL);
+	      if (visible && mode == GTK_CELL_RENDERER_MODE_ACTIVATABLE)
+	        {
+		  if (gtk_cell_renderer_activate (info->cell,
+						  event,
+						  tree_column->tree_view,
+						  path_string,
+						  background_area,
+						  cell_area,
+						  flags))
+		    return TRUE;
+		}
+	      else if (visible && mode == GTK_CELL_RENDERER_MODE_EDITABLE)
+	        {
+		  *editable_widget =
+		    gtk_cell_renderer_start_editing (info->cell,
+						     event,
+						     tree_column->tree_view,
+						     path_string,
+						     background_area,
+						     cell_area,
+						     flags);
+
+		  if (*editable_widget != NULL)
+		    {
+		      g_return_val_if_fail (GTK_IS_CELL_EDITABLE (*editable_widget), FALSE);
+
+		      return TRUE;
+		    }
+		}
+	    }
+	}
       real_cell_area.x += (info->requested_width + tree_column->spacing);
     }
-  if (! render)
+
+  if (action == CELL_ACTION_FOCUS)
     {
       if (min_x >= max_x || min_y >= max_y)
 	{
 	  *focus_rectangle = *cell_area;
-	  focus_rectangle->x -= focus_line_width;
-	  focus_rectangle->y -= focus_line_width;
-	  focus_rectangle->width += 2 * focus_line_width;
-	  focus_rectangle->height += 2 * focus_line_width;
+	  /* don't change the focus_rectangle, just draw it nicely inside
+	   * the cell area */
 	}
       else
 	{
@@ -2380,6 +2611,8 @@ gtk_tree_view_column_cell_render_or_focus (GtkTreeViewColumn *tree_column,
 	  focus_rectangle->height = (max_y - min_y) + 2 * focus_line_width;
 	}
     }
+
+  return FALSE;
 }
 
 /**
@@ -2407,14 +2640,14 @@ _gtk_tree_view_column_cell_render (GtkTreeViewColumn *tree_column,
   g_return_if_fail (cell_area != NULL);
   g_return_if_fail (expose_area != NULL);
 
-  gtk_tree_view_column_cell_render_or_focus (tree_column,
-					     window,
-					     background_area,
-					     cell_area,
-					     expose_area,
-					     flags,
-					     TRUE,
-					     NULL);
+  gtk_tree_view_column_cell_process_action (tree_column,
+					    window,
+					    background_area,
+					    cell_area,
+					    flags,
+					    CELL_ACTION_RENDER,
+					    expose_area,
+					    NULL, NULL, NULL, NULL);
 }
 
 gboolean
@@ -2426,43 +2659,18 @@ _gtk_tree_view_column_cell_event (GtkTreeViewColumn  *tree_column,
 				  GdkRectangle       *cell_area,
 				  guint               flags)
 {
-  gboolean visible, mode;
-
   g_return_val_if_fail (GTK_IS_TREE_VIEW_COLUMN (tree_column), FALSE);
 
-  g_object_get (G_OBJECT (((GtkTreeViewColumnCellInfo *) tree_column->cell_list->data)->cell),
-		"visible", &visible,
-		"mode", &mode,
-		NULL);
-  if (visible && mode == GTK_CELL_RENDERER_MODE_ACTIVATABLE)
-    {
-      if (gtk_cell_renderer_activate (((GtkTreeViewColumnCellInfo *) tree_column->cell_list->data)->cell,
-				      event,
-				      tree_column->tree_view,
-				      path_string,
-				      background_area,
-				      cell_area,
-				      flags))
-	return TRUE;
-    }
-  else if (visible && mode == GTK_CELL_RENDERER_MODE_EDITABLE)
-    {
-      *editable_widget = gtk_cell_renderer_start_editing (((GtkTreeViewColumnCellInfo *) tree_column->cell_list->data)->cell,
-							  event,
-							  tree_column->tree_view,
-							  path_string,
-							  background_area,
-							  cell_area,
-							  flags);
-
-      if (*editable_widget != NULL)
-	{
-	  g_return_val_if_fail (GTK_IS_CELL_EDITABLE (*editable_widget), FALSE);
-
-	  return TRUE;
-	}
-    }
-  return FALSE;
+  return gtk_tree_view_column_cell_process_action (tree_column,
+						   NULL,
+						   background_area,
+						   cell_area,
+						   flags,
+						   CELL_ACTION_EVENT,
+						   NULL, NULL,
+						   editable_widget,
+						   event,
+						   path_string);
 }
 
 
@@ -2470,6 +2678,7 @@ gboolean
 _gtk_tree_view_column_cell_focus (GtkTreeViewColumn *tree_column,
 				  gint               direction)
 {
+  /* FIXME 73676: allow focusing individual cells */
   if (GTK_TREE_VIEW (tree_column->tree_view)->priv->focus_column == tree_column)
     return FALSE;
   return TRUE;
@@ -2509,14 +2718,15 @@ _gtk_tree_view_column_cell_draw_focus (GtkTreeViewColumn       *tree_column,
   else
     {
       GdkRectangle focus_rectangle;
-      gtk_tree_view_column_cell_render_or_focus (tree_column,
-						 window,
-						 background_area,
-						 cell_area,
-						 expose_area,
-						 flags,
-						 FALSE,
-						 &focus_rectangle);
+      gtk_tree_view_column_cell_process_action (tree_column,
+						window,
+						background_area,
+						cell_area,
+						flags,
+						CELL_ACTION_FOCUS,
+						expose_area,
+						&focus_rectangle,
+						NULL, NULL, NULL);
 
       cell_state = flags & GTK_CELL_RENDERER_SELECTED ? GTK_STATE_SELECTED :
 	      (flags & GTK_CELL_RENDERER_PRELIT ? GTK_STATE_PRELIGHT :
@@ -2549,6 +2759,8 @@ gtk_tree_view_column_cell_is_visible (GtkTreeViewColumn *tree_column)
 {
   GList *list;
 
+  g_return_val_if_fail (GTK_IS_TREE_VIEW_COLUMN (tree_column), FALSE);
+
   for (list = tree_column->cell_list; list; list = list->next)
     {
       GtkTreeViewColumnCellInfo *info = (GtkTreeViewColumnCellInfo *) list->data;
@@ -2604,3 +2816,55 @@ _gtk_tree_view_column_stop_editing (GtkTreeViewColumn *tree_column)
 
   tree_column->editable_widget = NULL;
 }
+
+void
+_gtk_tree_view_column_get_neighbor_sizes (GtkTreeViewColumn *column,
+					  GtkCellRenderer   *cell,
+					  gint              *left,
+					  gint              *right)
+{
+  GList *list;
+
+  if (left)
+    {
+      *left = 0;
+
+      for (list = column->cell_list; list; list = list->next)
+        {
+	  GtkTreeViewColumnCellInfo *info =
+	    (GtkTreeViewColumnCellInfo *)list->data;
+
+	  if (info->cell == cell)
+	    break;
+
+	  *left += info->real_width;
+	}
+    }
+
+  if (right)
+    {
+      *right = 0;
+
+      for (list = column->cell_list; list; list = list->next)
+        {
+	  GtkTreeViewColumnCellInfo *info =
+	    (GtkTreeViewColumnCellInfo *)list->data;
+
+	  if (info->cell == cell)
+	    break;
+	}
+
+      /* skip cell */
+      if (list && list->next)
+        {
+	  list = list->next;
+	  for ( ; list; list = list->next)
+	    {
+	      GtkTreeViewColumnCellInfo *info =
+	        (GtkTreeViewColumnCellInfo *)list->data;
+
+	      *right += info->real_width;
+	    }
+	}
+    }
+}
